/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef WMSTACKING_H_
#define WMSTACKING_H_

void wRaiseFrame(WCoreWindow *frame);

void wLowerFrame(WCoreWindow *frame);

void wRaiseLowerFrame(WCoreWindow *frame);

void AddToStackList(WCoreWindow *frame);

void RemoveFromStackList(WCoreWindow *frame);

void ChangeStackingLevel(WCoreWindow *frame, int new_level);

void RemakeStackList(WScreen *scr);

void CommitStacking(WScreen *scr);


#endif
