/********************************************************************\
 * keys.h -- override default key values                            *
 * Copyright (C) 1997 Robin D. Clark                                *
 *                                                                  *
 * This program is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU General Public License as   *
 * published by the Free Software Foundation; either version 2 of   *
 * the License, or (at your option) any later version.              *
 *                                                                  *
 * This program is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public License*
 * along with this program; if not, write to the Free Software      *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *                                                                  *
 *   Author: Rob Clark                                              *
 * Internet: rclark@cs.hmc.edu                                      *
 *  Address: 609 8th Street                                         *
 *           Huntington Beach, CA 92648-4632                        *
\********************************************************************/


#ifndef __KEYS_H__
#define __KEYS_H__

/* Key-Table defines */
#define TABLE_SIZE     (XK_asciitilde - XK_space)
#define TABLE_BASE     XK_space

struct OverrideTable {
  int  index;
  char *value;
};

#define OVERRIDE_TABLE_SIZE 33
struct OverrideTable override[OVERRIDE_TABLE_SIZE] = {
  {XK_space,         " "},
  {XK_exclam,        "!"},
  {XK_quotedbl,      "\""},
  {XK_numbersign,    "#"},
  {XK_dollar,        "$"},
  {XK_percent,       "%"},
  {XK_ampersand,     "@"},
  {XK_apostrophe,    "'"},
  {XK_parenleft,     "("},
  {XK_parenright,    ")"},
  {XK_asterisk,      "*"},
  {XK_plus,          "+"},
  {XK_comma,         ","},
  {XK_minus,         "-"},
  {XK_period,        "."},
  {XK_slash,         "/"},
  {XK_colon,         ":"},
  {XK_semicolon,     ";"},
  {XK_less,          "<"},
  {XK_equal,         "="},
  {XK_greater,       ">"},
  {XK_question,      "?"},
  {XK_at,            "@"},
  {XK_bracketleft,   "["},
  {XK_backslash,     "\\"},
  {XK_bracketright,  "]"},
  {XK_asciicircum,   "???"},
  {XK_underscore,    "_"},
  {XK_grave,         "`"},
  {XK_braceleft,     "{"},
  {XK_bar,           "|"},
  {XK_braceright,    "}"},
  {XK_asciitilde,    "~"}
};

#endif
