#ifdef XPM
static char *BUTTON_ARROW_XPM[] = {
"    15    10       4           1",
/* colors */
"  c None",
". c #000000",
"# c #ffffff",
"a c #bebebe",
/* pixels */
"          ....a",
"    ..    .   #",
"   . a    .   #",
"  .  a.....   #",
" .            #",
".             #",
" .            #",
"  .  ##########",
"   . #         ",
"    .#         "
};



/* button pixmaps */
/* the first one is for normal state, the second is for when
 * the button is pushed */
#ifndef NEWSTYLE
static char *PRED_CLOSE_XPM[] = {
"20 10 3 1",
" 	c #000000",
".	c #616561",
"X	c None",
" .XXXXXX. ..XXXXXX..",
". .XXXX. ....XXXX...",
"X. .XX. .XX...XX...X",
"XX. .. .XXXX......XX",
"XXX.  .XXXXXX....XXX",
"XXX.  .XXXXXX....XXX",
"XX. .. .XXXX......XX",
"X. .XX. .XX...XX...X",
". .XXXX. ....XXXX...",
" .XXXXXX. ..XXXXXX.."};


static char *PRED_BROKENCLOSE_XPM[] = {
"20 10 4 1",
" 	c #000000",
".	c #616561",
"-     	c #000000",      
"X	c None",
" .XXXXXX. --XXXXXX--",
". .XXXX. .---XXXX---",
"X. XXXX .XX--XXXX--X",
"XXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXX",
"XXXXXXXXXXXXXXXXXXXX",
"X. XXXX .XX--XXXX--X",
". .XXXX. .---XXXX---",
" .XXXXXX. --XXXXXX--"};



static char *PRED_ICONIFY_XPM[] = {
"    20    10        3            1",
". c #000000",
"# c None",
"o c #616561",
"..........oooooooooo",
"..........oooooooooo",
"..........oooooooooo",
".########.o########o",
".########.o########o",
".########.o########o",
".########.o########o",
".########.o########o",
".########.o########o",
"..........oooooooooo"
};
#else /* NEWSTYLE */
#define PRED_ICONIFY_XBM_WIDTH 20
#define PRED_ICONIFY_XBM_HEIGHT 10
static unsigned char PRED_ICONIFY_XBM_DATA[] = {
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x01,0x06,0xf8,0x01,0x06,0xf8,
 0x01,0x06,0xf8,0x01,0x06,0xf8,0x01,0x06,0xf8,0x01,0x06,0xf8,0xff,0xff,0xff
 };

#define PRED_BROKENCLOSE_XBM_WIDTH 20
#define PRED_BROKENCLOSE_XBM_HEIGHT 10
static unsigned char PRED_BROKENCLOSE_XBM_DATA[] = {
 0x03,0x0f,0xfc,0x87,0x1f,0xfe,0x86,0x19,0xf6,0x00,0x00,0xf0,0x00,0x00,0xf0,
 0x00,0x00,0xf0,0x00,0x00,0xf0,0x86,0x19,0xf6,0x87,0x1f,0xfe,0x03,0x0f,0xfc
 };

#define PRED_CLOSE_XBM_WIDTH 20
#define PRED_CLOSE_XBM_HEIGHT 10
static unsigned char PRED_CLOSE_XBM_DATA[] = {
 0x03,0x0f,0xfc,0x87,0x1f,0xfe,0xce,0x39,0xf7,0xfc,0xf0,0xf3,0x78,0xe0,0xf1,
 0x78,0xe0,0xf1,0xfc,0xf0,0xf3,0xce,0x39,0xf7,0x87,0x1f,0xfe,0x03,0x0f,0xfc
 };
#endif /* NEWSTYLE */

#else /* XPM */
/* bitmap equivalents of the predefined pixmaps */
#define BUTTON_ARROW_XBM_WIDTH 15
#define BUTTON_ARROW_XBM_HEIGHT 10
static unsigned char BUTTON_ARROW_XBM_DATA[] = {
   0x00, 0xfc, 0x30, 0x84, 0x28, 0x84, 0xe4, 0x87, 0x02, 0x80, 0x01, 0x80,
   0x02, 0x80, 0x04, 0x80, 0x08, 0x80, 0x30, 0x80};
static unsigned char BUTTON_ARROW_MASK_XBM_DATA[] = {
   0x00, 0xfc, 0x30, 0xc4, 0x28, 0xc4, 0xe4, 0xc7, 0x02, 0xc0, 0x01, 0xc0,
   0x02, 0xc0, 0xe4, 0xff, 0x28, 0x80, 0x30, 0x80};


#define PRED_CLOSE_XBM_WIDTH 20
#define PRED_CLOSE_XBM_HEIGHT 10
static unsigned char PRED_CLOSE_XBM_DATA[] = {
   0x03, 0x0f, 0xfc, 0x87, 0x0f, 0xfc, 0xce, 0x01, 0xf0, 0xfc, 0x00, 0xf0,
   0x70, 0x00, 0xf0, 0xf8, 0x00, 0xf0, 0xdc, 0x01, 0xf0, 0x8e, 0x03, 0xf0,
   0x07, 0x0f, 0xfc, 0x03, 0x0e, 0xfc};


#define PRED_BROKEN_CLOSE_XBM_WIDTH 20
#define PRED_BROKEN_CLOSE_XBM_HEIGHT 10
static unsigned char PRED_BROKEN_CLOSE_XBM_DATA[] = {
   0x03, 0x0f, 0xfc, 0x87, 0x17, 0xfa, 0x86, 0x19, 0xf6, 0x00, 0x00, 0xf0,
   0x00, 0x00, 0xf0, 0x00, 0x00, 0xf0, 0x00, 0x00, 0xf0, 0x86, 0x19, 0xf6,
   0x87, 0x17, 0xfa, 0x03, 0x0f, 0xfc};


#define PRED_ICONIFY_XBM_WIDTH 20
#define PRED_ICONIFY_XBM_HEIGHT 10
static unsigned char PRED_ICONIFY_XBM_DATA[] = {
 0xff,0x03,0xf0,0xff,0x03,0xf0,0xff,0x03,0xf0,0x01,0x02,0xf0,0x01,0xfe,0xf0,
 0x01,0xfe,0xf0,0x01,0x86,0xf0,0x01,0x86,0xf0,0x01,0x86,0xf0,0xff,0xff,0xf0
 };


#endif /* !XPM */
