/* 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

typedef union {
    int num;
    char *ptr;
    double real;
} YYSTYPE;

#define UDOMAIN		0x100
#define UKEY		0x101
#define UVALUE		0x102
#define UTYPE		0x103
#define UVERSION	0x104

#define TSTRING		0x110
#define TBOOLEAN 	0x111
#define TINTEGER 	0x112
#define TDOUBLE		0x113
#define TSTRINGLIST	0x114

#define DOUBLE		0x120
#define INTEGER		0x121
#define STRING		0x122
#define BOOLEAN		0x123
#define KEYWORD		0x124

extern YYSTYPE UDDBlval;
