
#ifndef _WAPPMENU_H_
#define _WAPPMENU_H_

WMenu *wAppMenuGet(WScreen *scr, Window window);
void wAppMenuDestroy(WMenu *menu);

void wAppMenuMap(WMenu *menu);
void wAppMenuUnmap(WMenu *menu);


#endif
