/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef WMDIALOG_H_
#define WMDIALOG_H_

#define WD_INFORMATION 	0
#define WD_ERROR	1
#define WD_CONFIRMATION	2
#define WD_EXIT_CONFIRM 3
#define WD_INPUT	4

#define WDB_OK		(1<<0)
#define WDB_CANCEL	(1<<1)
#define WDB_YES		(1<<2)
#define WDB_NO		(1<<3)
#define WDB_EXIT	(1<<4)

#include "text.h"

typedef struct WDialog {
    WWindow *wwin;
    WCoreWindow *core;
    WButton *but1, *but2;    
    WTextInput   *wText;
    int selected;
    int type;
    int text_h;
    char *message;
    char *title;
    WPixmap *pixmap;
    GC gc;
} WDialog;



void wDialogInitialize();
int wMessageDialog(WScreen *scr, char *title, char *message, int type);
void wDialogPaint(WDialog *dlg);

#ifdef USE_ARG_DIALOG
char *wInputDialog( WScreen *src, char *title, char *message, char *value);
#endif

#endif
