/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef WMICON_H_
#define WMICON_H_

#include "wcore.h"
#include "window.h"
typedef struct WIcon {
    WCoreWindow *core;
    Window icon_win;		       /* client suplied icon window */
    char *icon_name;		       /* the icon name hint */
    unsigned int client_window:1;      /* icon window supplied by client */
    unsigned int client_pixmap:1;      /* pixmap supplied by client */
    unsigned int our_pixmap:1;
    unsigned int show_title:1;
    unsigned int has_titlebar:1;
    unsigned int force_paint:1;
    unsigned int use_our_pixmap:1;     /* always use user specified pixmap  */
    char *icon_file;
    char *icon_text;
    WWindow *owner;		       /* owner window */
    Pixmap pixmap;		       /* pixmap */
    Pixmap mask;		       /* mask pixmap */
#ifdef METROBUG
    Pixmap pix_image;
#endif
} WIcon;


WIcon *wIconCreateWithImage(WScreen *scr, char *icon_file);
WIcon *wIconCreate(WWindow *wwin);
void wIconDestroy(WIcon *icon);
void wIconPaint(WIcon *icon);
void wIconUpdate(WIcon *icon);
#endif
