#ifdef XPM
static char * GNUSTEP_XPM[] = {
"45 45 8 1",
" 	c None",
".	c #AEBAAAAAB6DA",
"X	c #9A699A699E79",
"o	c #596559656185",
"O	c #000000000000",
"+	c #3CF33CF33CF3",
"@	c #CF3CCF3CCF3C",
"#	c #FFFFFFFFFFFF",
"                                             ",
"                .Xooooooooo.                 ",
"            ..ooOOOOOOOOOOOOOo.              ",
"          .XoOOOOOOOOOOOOOOOOOO+             ",
"        ..oOOOOOOOOOOOOOOOOOOOOOO+           ",
"       .XOOOOOOOOOOOOOOOOOOOOOOOOOOo         ",
"       oOOOOOOOOOOOOOOOOOOOOOOOOOOOO+        ",
"     .oOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO+ooooooo",
"    .+OOOOOOOOOOOOOOOOOOOOOOOOoXXXXX.XOOOOOOO",
"   .XOOOOOOOOOOOOOOOOOOOOOOOOOX#######.OOOOOO",
"   XOOOOOOOOOOOOOOOOOOOOOOOOOOX########oOOOOO",
"   OOOOOOOOOOOOOOOOOOOOOOOOOOOX#########oOOOO",
"  oOOOOOOOOOOOOOOOOOOOOOOOOOOOX#########@OOOO",
"  OOOOOOOOOOOOOOOOOOOOOOOOOOOOX##########oOOO",
" oOOOOOOOOOOOOOOOOOOOOOOOOOOOOX##########@OOO",
" OOOOOOOOOOOOOOOOOOOOOOOOOOOOOX###########+OO",
"XOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX###########oOO",
"oOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX###########@OO",
"+OOOOOOOOOOOOOOOOOOOOOOOOOOOOOX###########@OO",
"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX############XO",
"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX############XO",
"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX############XO",
"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOX############XO",
"OOOOOOOOOOOOOOOoX.X.X.X.XX.XXX@############XO",
"OOOOOOOOOOOOOOOX###########################XO",
"OOOOOOOOOOOOOOOX###########################XO",
"OOOOOOOOOOOOOOOX###########################XO",
"+OOOOOOOOOOOOOOX###########################OO",
"oOOOOOOOOOOOOOOX###########################OO",
"XOOOOOOOOOOOOOOX##########################XOO",
".OOOOOOOOOOOOOOX##########################oOO",
".oOOOOOOOOOOOOOX#########################@OOO",
"..OOOOOOOOOOOOOX#########################+OOO",
"X.oOOOOOOOOOOOOX########################@OOOO",
"..XOOOOOOOOOOOOX########################oOOOO",
"X.XXOOOOOOOOOOOX#######################oOOOOO",
".XXXXOOOOOOOOOOX######################XOOOOOO",
"OOOOOOOXXXXXXXX@#####################@OOOOOOO",
"OOOOOOOo############################@OOOOOOOO",
"OOOOOOOOO@#########################oOOOOOOOOO",
"OOOOOOOOOOX######################@oOOOOOOOOOO",
"OOOOOOOOOOOOX###################XOOOOOOOOOOOO",
"OOOOOOOOOOOOOOX@#############@XOOOOOOOOOOOOOO",
"OOOOOOOOOOOOOOOOOXXX#####XXXOOOOOOOOOOOOOOOOO",
"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO"};
#else /* XPM */
#define GNUSTEP_XBM_WIDTH 45
#define GNUSTEP_XBM_HEIGHT 45
/* GNUstep logo bitmap */
static unsigned char GNUSTEP_XBM_DATA[] = {
 0x00,0x00,0x00,0x00,0x00,0xe0,0x00,0x00,0xfc,0x07,0x00,0xe0,0x00,0xc0,0xff,
 0x3f,0x00,0xe0,0x00,0xf0,0xff,0xff,0x00,0xe0,0x00,0xfc,0xff,0xff,0x03,0xe0,
 0x00,0xfe,0xff,0xff,0x0f,0xe0,0x80,0xff,0xff,0xff,0x1f,0xe0,0xc0,0xff,0xff,
 0xff,0xff,0xff,0xe0,0xff,0xff,0x7f,0xc0,0xff,0xe0,0xff,0xff,0x3f,0x80,0xff,
 0xf0,0xff,0xff,0x3f,0x80,0xff,0xf8,0xff,0xff,0x3f,0x00,0xff,0xfc,0xff,0xff,
 0x3f,0x00,0xfe,0xfc,0xff,0xff,0x3f,0x00,0xfe,0xfe,0xff,0xff,0x3f,0x00,0xfc,
 0xfe,0xff,0xff,0x3f,0x00,0xfc,0xfe,0xff,0xff,0x3f,0x00,0xfc,0xff,0xff,0xff,
 0x3f,0x00,0xf8,0xff,0xff,0xff,0x3f,0x00,0xf8,0xff,0xff,0xff,0x3f,0x00,0xf0,
 0xff,0xff,0xff,0x3f,0x00,0xf0,0xff,0xff,0xff,0x3f,0x00,0xf0,0xff,0xff,0xff,
 0x3f,0x00,0xf0,0xff,0xff,0x00,0x00,0x00,0xf0,0xff,0x7f,0x00,0x00,0x00,0xf0,
 0xff,0x7f,0x00,0x00,0x00,0xf0,0xff,0x7f,0x00,0x00,0x00,0xf0,0xff,0x7f,0x00,
 0x00,0x00,0xf8,0xff,0x7f,0x00,0x00,0x00,0xf8,0xfe,0x7f,0x00,0x00,0x00,0xf8,
 0xfe,0x7f,0x00,0x00,0x00,0xfc,0xfe,0x7f,0x00,0x00,0x00,0xfc,0xfc,0x7f,0x00,
 0x00,0x00,0xfe,0xfc,0x7f,0x00,0x00,0x00,0xfe,0xf8,0x7f,0x00,0x00,0x00,0xff,
 0xf0,0x7f,0x00,0x00,0x80,0xff,0xe0,0x7f,0x00,0x00,0x80,0xff,0x7f,0x00,0x00,
 0x00,0xc0,0xff,0xff,0x00,0x00,0x00,0xe0,0xff,0xff,0x01,0x00,0x00,0xf8,0xff,
 0xff,0x03,0x00,0x00,0xfc,0xff,0xff,0x0f,0x00,0x00,0xfe,0xff,0xff,0x3f,0x00,
 0x80,0xff,0xff,0xff,0xff,0x01,0xf0,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff
 };

/* mask for GNUstep logo bitmap */
static unsigned char GNUSTEP_MASK_XBM_DATA[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x07, 0x00, 0x00,
   0x00, 0xc0, 0xff, 0x3f, 0x00, 0x00, 0x00, 0xf0, 0xff, 0xff, 0x00, 0x00,
   0x00, 0xfc, 0xff, 0xff, 0x03, 0x00, 0x00, 0xfe, 0xff, 0xff, 0x0f, 0x00,
   0x80, 0xff, 0xff, 0xff, 0x1f, 0x00, 0xc0, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xe0, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xe0, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xf0, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xf8, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xfc, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfc, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfe, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xfc, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfc, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xf8, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xf0, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xe0, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f,
   0xff, 0xff, 0xff, 0xff, 0xff, 0x1f};
#endif /* !XPM */
