/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* anywhere */
#define WKBD_ROOTMENU		1
/* window */
#define WKBD_MINIATURIZE	3
#define WKBD_HIDE		4
#define WKBD_MAXIMIZE		5

#define WKBD_SELECTWINDOW	6

#define WKBD_RESTORESIZE	10
/* window, menu */
#define WKBD_RAISE		11
#define WKBD_LOWER		12
#define WKBD_RAISELOWER		13
/* window */
#define WKBD_SHADE		14
/* window, menu */
#define WKBD_CLOSE		15
/* window */
#define WKBD_FOCUSNEXT		16
#define WKBD_FOCUSPREV		17
/* anywhere */
#define WKBD_NEXTMENU		18
#define WKBD_PREVMENU		19

#define WKBD_WORKSPACE1		20
#define WKBD_WORKSPACE2		21
#define WKBD_WORKSPACE3		22
#define WKBD_WORKSPACE4		23
#define WKBD_WORKSPACE5		24
#define WKBD_WORKSPACE6		25
#define WKBD_WORKSPACE7		26
#define WKBD_WORKSPACE8		27
#define WKBD_WORKSPACE9		28
#define WKBD_NEXTWORKSPACE	29
#define WKBD_PREVWORKSPACE	30

#define WKBD_DOWN		31
#define WKBD_RIGHT		32
#define WKBD_LEFT		33
#define WKBD_UP			34

#define WKBD_LAST		35


typedef struct WShortKey {
    unsigned int modifier;
    unsigned int keycode;
    struct WShortKey *alternate;
} WShortKey;

void wKeyboardInitialize();
